import unittest
import unittest.mock

# Potential improvement: split testcases with different fixtures needed into their own thing (make_HParsedToken, read_member, read_AST_not_null)
# TODO: reduce boilerplate using setUp(), tearDown() and decorators for testcases with multiple test functions
class HParseResultTests(unittest.TestCase):
	#def test_valid_address(self):
	#	raise ValueError

	def test_nullpointer(self):
		with self.assertRaises(ValueError):
			res = HParseResult(0)

	def test_make_HParsedToken(self):
		gdbv_patcher = unittest.mock.patch('gdb.Value', autospec=True)
		hpt_patcher = unittest.mock.patch('__main__.HParsedToken', autospec=True)
		hpr_patcher = unittest.mock.patch.object(HParseResult, 'read_AST_not_null', return_value=True)
		rm_patcher = unittest.mock.patch.object(HParseResult, 'read_member', return_value=0xdeadfff0)
		gdbv_mock_object = gdbv_patcher.start()
		hpt_mock_object = hpt_patcher.start()
		hpr_patcher.start()
		rm_mock_object = rm_patcher.start()
		hpr = HParseResult(0xdeadbeef)
		tok = hpr.make_HParsedToken()
		rm_patcher.stop()
		hpr_patcher.stop()
		hpt_patcher.stop()
		gdbv_patcher.stop()
		# Check that read_member('ast') is called
		self.assertEqual(rm_mock_object.call_args, unittest.mock.call('ast'))
		# Check that HParsedToken() is called with the value returned by read_member()
		self.assertEqual(hpt_mock_object.call_args, unittest.mock.call(0xdeadfff0))

	def test_make_HParsedToken_if_no_ast(self):
		gdbv_patcher = unittest.mock.patch('gdb.Value', autospec=True)
		hpt_patcher = unittest.mock.patch('__main__.HParsedToken', autospec=True)
		hpr_patcher = unittest.mock.patch.object(HParseResult, 'read_AST_not_null', return_value=False)
		gdbv_patcher.start()
		hpt_patcher.start()
		hpr_patcher.start()
		hpr = HParseResult(0xdeadbeef)
		tok = hpr.make_HParsedToken()
		hpr_patcher.stop()
		hpt_patcher.stop()
		gdbv_patcher.stop()
		self.assertIsNone(tok)

	def test_read_member(self):
		hpr_pointer_type_patcher = unittest.mock.patch.object(HParseResult, 'HParseResult_t_p', spec=gdb.Type)
		gdbv_patcher = unittest.mock.patch('gdb.Value', autospec=True)
		gdb_lookup_type_patcher = unittest.mock.patch.object(gdb, 'lookup_type')
		init_patcher = unittest.mock.patch.object(HParseResult, '__init__', return_value=None)

		hpr_pointer_type_patcher.start()
		gdbv_mock_object = gdbv_patcher.start()
		gdb_lookup_type_mock_object = gdb_lookup_type_patcher.start()
		init_patcher.start()

		result = HParseResult(0xdeadbeef)
		result.address = 0xdeadbeef
		member = result.read_member('ast')

		init_patcher.stop()
		gdb_lookup_type_patcher.stop()
		gdbv_patcher.stop()
		hpr_pointer_type_patcher.stop()

		# Check that struct being indexed is at self.address
		self.assertEqual(gdbv_mock_object.mock_calls[0], unittest.mock.call(0xdeadbeef))
		# Check that the value returned is parse_result_gdb_value['foo']
		self.assertEqual(gdbv_mock_object.mock_calls[-1], unittest.mock.call().cast().__getitem__('ast'))
		# Check that gdb.lookup_type is not called
		self.assertEqual(gdb_lookup_type_mock_object.mock_calls, [])

	def test_read_member_with_type_lookup(self):
		gdb_lookup_type_patcher = unittest.mock.patch.object(gdb, 'lookup_type')
		gdbv_patcher = unittest.mock.patch('gdb.Value', autospec=True)
		init_patcher = unittest.mock.patch.object(HParseResult, '__init__', return_value=None)
		# This is cached in the class the first time read_member() or read_AST_not_null() is called
		# To test the case where it's not yet cached, we explicitly set it to None
		HParseResult.HParseResult_t_p = None

		gdb_lookup_type_mock_object = gdb_lookup_type_patcher.start()
		gdbv_mock_object = gdbv_patcher.start()
		init_patcher.start()

		result = HParseResult(0xdeadbeef)
		result.address = 0xdeadbeef
		member = result.read_member('ast')

		init_patcher.stop()
		gdbv_patcher.stop()
		gdb_lookup_type_patcher.stop()

		# Check that HParseResult* has been looked up
		self.assertEqual(gdb_lookup_type_mock_object.mock_calls, [unittest.mock.call('HParseResult'), unittest.mock.call().pointer()])
		# Check that the struct being indexed is at self.address
		self.assertEqual(gdbv_mock_object.mock_calls[0], unittest.mock.call(0xdeadbeef))
		# Check that the value returned is parse_result_gdb_value['foo']
		self.assertEqual(gdbv_mock_object.mock_calls[-1], unittest.mock.call().cast().__getitem__('ast'))

	@unittest.skip("TODO: param is validated by gdb.Value, possibly unnecessary")
	def test_read_member_invalid_param(self):
		hpr_pointer_type_patcher = unittest.mock.patch.object(HParseResult, 'HParseResult_t_p', spec=gdb.Type)
		gdbv_patcher = unittest.mock.patch('gdb.Value', autospec=True)
		gdb_lookup_type_patcher = unittest.mock.patch.object(gdb, 'lookup_type')
		init_patcher = unittest.mock.patch.object(HParseResult, '__init__', return_value=None)

		hpr_pointer_type_patcher.start()
		gdbv_mock_object = gdbv_patcher.start()
		gdb_lookup_type_mock_object = gdb_lookup_type_patcher.start()
		init_patcher.start()

		with self.assertRaises(gdb.error):
			result = HParseResult(0xdeadbeef)
			result.address = 0xdeadbeef
			member = result.read_member('invalid_member_value')

			init_patcher.stop()
			gdb_lookup_type_patcher.stop()
			gdbv_patcher.stop()
			hpr_pointer_type_patcher.stop()

#class HParsedTokenCreation(unittest.TestCase):

class HCountedArrayCreation(unittest.TestCase):
	#TODO: read_member return value should be mock gdb.Value
	@unittest.mock.patch.object(HCountedArray, 'read_member', return_value=0xdeadbccc)
	@unittest.mock.patch('gdb.Value', autospec=True)
	def test_hca_creation(self, gdbv_mock, hca_readmember_mock):
		hca = HCountedArray(0xdeadbeef)
		self.assertEqual(hca.address, 0xdeadbeef)
		self.assertEqual(hca_readmember_mock.mock_calls, [unittest.mock.call('capacity'), unittest.mock.call('used'), unittest.mock.call('arena'), unittest.mock.call('elements')])

class HCountedArrayReadMember(unittest.TestCase):
	@unittest.mock.patch.object(HCountedArray, '__init__', return_value=None)
	@unittest.mock.patch.object(gdb, 'lookup_type', autospec=True)
	@unittest.mock.patch('gdb.Value', autospec=True)
	def test_read_member_with_type_lookup(self, gdbv_mock, gdb_lookup_mock, hca_init_mock):
		hca = HCountedArray(0xdeadbeef)
		hca.address = 0xdeadbeef
		HCountedArray.HCountedArray_t_p = None
		member = hca.read_member('capacity')

		self.assertEqual(gdb_lookup_mock.mock_calls, [unittest.mock.call("HCountedArray"), unittest.mock.call().pointer()])
		self.assertEqual(gdbv_mock.mock_calls, [unittest.mock.call(hca.address), unittest.mock.call().cast(HCountedArray.HCountedArray_t_p), unittest.mock.call().cast().__getitem__('capacity')])

	@unittest.mock.patch.object(HCountedArray, 'HCountedArray_t_p', spec=gdb.Type)
	@unittest.mock.patch.object(HCountedArray, '__init__', return_value=None)
	@unittest.mock.patch.object(gdb, 'lookup_type', autospec=True)
	@unittest.mock.patch('gdb.Value', autospec=True)
	def test_read_member_no_type_lookup(self, gdbv_mock, gdb_lookup_mock, hca_init_mock, hca_t_p_mock):
		hca = HCountedArray(0xdeadbeef)
		hca.address = 0xdeadbeef
		member = hca.read_member('capacity')

		self.assertEqual(gdbv_mock.mock_calls, [unittest.mock.call(hca.address), unittest.mock.call().cast(HCountedArray.HCountedArray_t_p), unittest.mock.call().cast().__getitem__('capacity')])
		self.assertEqual(gdb_lookup_mock.mock_calls, [])

class HBytesCreation(unittest.TestCase):
	@unittest.mock.patch('gdb.Value', autospec=True)
	def test_hb_creation(self, gdbv_mock):
		hb = HBytes(gdb.Value(0xdeadbeef))
		self.assertEqual(gdbv_mock.mock_calls, [unittest.mock.call(0xdeadbeef), unittest.mock.call().__getitem__('len'), unittest.mock.call().__getitem__('token')])

class HBytesStr(unittest.TestCase):
	@unittest.mock.patch('gdb.Value', autospec=True)
	def test_str(self, gdbv_mock):
		len_mock = unittest.mock.MagicMock()
		len_mock.__eq__.return_value = False
		len_mock.__str__.return_value = "4"

		token_mock = unittest.mock.MagicMock()
		token_mock.string.return_value = 'test'
		hb = HBytes(gdb.Value(0xdeadbeef))
		hb.len = len_mock
		hb.token = token_mock
		teststr = hb.__str__()
		self.assertEqual(teststr, "{ token: \"test\", len: 4 }")
		self.assertEqual(len_mock.mock_calls, [unittest.mock.call.__eq__(0),  unittest.mock._Call(("__str__", (), {}))]) # using _Call() here is a workaround to .__str__() just converting call() to a string when using the call() shorthand
		self.assertEqual(token_mock.mock_calls, [unittest.mock.call.string("UTF-8", "replace", len_mock)])

	@unittest.mock.patch('gdb.Value', autospec=True)
	def test_str_len_0(self, gdbv_mock):
		len_mock = unittest.mock.MagicMock()
		len_mock.__eq__.return_value = True

		token_mock = unittest.mock.MagicMock()
		hb = HBytes(gdb.Value(0xdeadbeef))
		hb.len = len_mock
		hb.token = token_mock
		teststr = hb.__str__()
		self.assertEqual(teststr, "{ token: \"\", len: 0 }")
		self.assertNotIn(unittest.mock.call().__str__(), len_mock.mock_calls)
		self.assertNotIn(unittest.mock.call().__str__(), token_mock.mock_calls)


class ASTManagerSetTopNode(unittest.TestCase):
	def setUp(self):
		self.hpr_patcher = unittest.mock.patch('__main__.HParseResult', autospec=True, return_value=unittest.mock.sentinel.hparseresult)
		self.parser_patcher = unittest.mock.patch('__main__.Parser', autospec=True)
		self.hpr_mock_object = self.hpr_patcher.start()
		self.parser_mock_object = self.parser_patcher.start()

	def tearDown(self):
		self.parser_patcher.stop()
		self.hpr_patcher.stop()

	def test_set_top_node(self):
		p = Parser('test_parser', 0xdeadbfff)
		ast_manager = ASTManager()
		ast_manager.set_top_node(0xdeadbeef, p)

		self.assertEqual(ast_manager.top_node, unittest.mock.sentinel.hparseresult)
		self.assertEqual(ast_manager.parser, p)
		self.assertEqual(self.hpr_mock_object.mock_calls, [unittest.mock.call(0xdeadbeef)])

	def test_set_top_node_none(self):
		ast_manager = ASTManager()
		ast_manager.set_top_node(None, None)

		self.assertIsNone(ast_manager.parser)
		self.assertIsNone(ast_manager.top_node)
		self.assertEqual(self.hpr_mock_object.mock_calls, [])

	def test_set_top_node_zero_address(self):
		ast_manager = ASTManager()
		ast_manager.set_top_node(0, None)

		self.assertIsNone(ast_manager.parser)
		self.assertIsNone(ast_manager.top_node)
		self.assertEqual(self.hpr_mock_object.mock_calls, [])

class ASTManagerPrintAST(unittest.TestCase):
	def setUp(self):
		self.hpr_patcher = unittest.mock.patch('__main__.HParseResult', autospec=True)
		self.parser_patcher = unittest.mock.patch('__main__.Parser', autospec=True)
		self.hpr_mock_object = self.hpr_patcher.start()
		self.parser_mock_object = self.parser_patcher.start()

	def tearDown(self):
		self.parser_patcher.stop()
		self.hpr_patcher.stop()

	@unittest.mock.patch("builtins.print")
	def test_print_ast(self, print_mock):
		p = Parser('test_parser', 0xdeadbfff)
		ast_manager = ASTManager()
		ast_manager.set_top_node(0xdeadbeef, p)

		ast_manager.print_ast()

		self.assertEqual(print_mock.mock_calls, [unittest.mock.call(ast_manager.parser), unittest.mock.call(ast_manager.top_node)])
