#import cProfile #DEBUG

class TopLevelParse:
	def __init__(self):
		self.parser_stacks = []
		self.parser_objs = {}
		self.index_tree_stack = []
		self.current_parse_depth = 0 # Used by the AST handling code.
		self.starting_input_index = []
		self.input_token_map = {}
		self.show_input_map_after_apply = False
		self.input_ptr = None
		# Technically, 1 virtual parse function + 1 h_do_parse + 1 perform_lowlevel_parse (if packrat)
		# TODO: len(self.parser_stacks[-1]) and self.current_parse_depth serve the same purpose. avoid multiple sources of truth
		# TODO: stack of trees to push/pop when switching parser stacks

		self.unclaimed_mem_use = 0
		# Holds 32 characters starting at state->input_stream[index], used by the GUI
		self.current_input_chunk = ''
		self.current_parser_env = ''
		# We save/push in perform_lowlevel_parse, but this is used to display them ahead of time
		self.h_do_parse_parser = None
		self.vt_types = None
		self.parser_decombinator = None

		# Counts stack pushes/pops to determine if stopping is needed for hammer-parse-after-apply
		self.ast_stack_index = None
		# Holds a reference to the relevant parser stack
		self.ast_selected_stack = None
		# Holds parse states down the stack from when hammer-parse-apply or hammer-parse-step-to-result was executed
		self.printing_tokens = [] # Holds the tokens yet to be printed by print_input_chunk as it walks the AST
		self.printing_index = 0 # Tokens have been printed up to this position in the array
		self.tokens_from_previous_chunk = []

	def init_parser(self):
		self.vt_types = VTTypes()
		self.parser_decombinator = ParserDecombinator(self.vt_types)

	# Called from h_packrat_parse()'s handler, where the parse state and arena get initialized
	def enter_h_packrat_parse(self, parser):
		parser_stack = ParserStack(None, None)
		self.parser_stacks.append(parser_stack)
		self.current_parse_depth = -1 # entering h_do_parse increments this by 1. To have it 0-indexed, it needs to be initialized to -1
		return 0

	# TODO: arena parameter is useless
	def enter_h_do_parse(self, parse_state, arena, parser):
		parser_stack = self.peek_parserstack()
		try:
			parser_obj = self.parser_objs[parser]
		except KeyError:
			parser_obj = Parser(None, parser)
			self.parser_objs[parser] = parser_obj
		self.h_do_parse_parser = parser_obj # TODO: current_parser_env should be set here instead too
		if parser_stack.parse_state is None and parser_stack.parse_state != parse_state:
			self.first_h_do_parse_after_packrat_parse(parse_state, arena)
		self.current_parse_depth += 1
		start_index = int(parse_state['input_stream']['index'])
		self.begin_new_token(start_index)

	def return_from_h_do_parse(self, parse_state, parser, ret_val):
		parser_stack = self.peek_parserstack()
		parser_obj = self.parser_by_address(parser) # We don't try to make a new Parser() object here, because it has to have been done in the corresponding enter_h_do_parse() call
		# The rationale for handling the "current" parser separately from stack:
		# Parsers are only pushed on the stack in perform_lowlevel_parse, which doesn't get called when h_do_parse() gets the result from cache.
		# If other backends are supported, this might change to pushing/popping the stack in h_do_parse()
		self.h_do_parse_parser = parser_obj # Restore the "current" parser, otherwise it'll show the parser h_do_parse() was last called with on the GUI and backtrace

		if ret_val:
			# We'd have to subtract 1 from len(parser_stack), but since we push/pop the stack in perform_lowlevel_parse, by this point it's 1 item shorter than the depth we're at
			#token_bounds = self.input_stream_indices[len(parser_stack.p_stack)] # TODO: ParserStack.get_stack_size() or similar
			#final_token_bounds = (token_bounds[0], int(parse_state['input_stream']['index']))
			#self.input_stream_indices[len(parser_stack.p_stack)] = final_token_bounds
			end_index = int(parse_state['input_stream']['index'])
			self.finish_token(end_index, ret_val)
		else:
			# If ret_val is NULL, the parse has failed. We can be sure the (start, end) tuple won't be needed here.
			self.discard_pending_tokens()
		self.current_parse_depth -= 1

	# Called from h_do_parse()'s handler, at which point we know the addresses of the state and arena
	def first_h_do_parse_after_packrat_parse(self, parse_state, arena):
		parser_stack = self.peek_parserstack()
		parser_stack.set_state(parse_state)

	# Popping the stack of stack of parsers
	def return_from_h_packrat_parse(self):
		old_stack = self.parser_stacks.pop()
		if old_stack.depth() > 0:
			print("Warning: parser stack not empty but parse is successful?")

		if not self.parser_stacks: # if self.parser_stacks is now empty, we've returned from h_packrat_parse() called by the "outermost" h_parse()
			self.current_parse_depth = -1 # Restore parse depth
		else:
			self.current_parse_depth = len(self.peek_parserstack().p_stack)-1 # This happens when a h_parse called from a semantic action or HContinuation returns (e. g. filters in pdf)

	# Memoize the parser object for this particular address, then push it on the stack
	# Returns the parser object we just initalized (or the one already existing)
	#TODO: memoize_parser method
	def enter_perform_lowlevel_parse(self, parser_addr):
		try:
			parser_obj = self.parser_objs[parser_addr]
		except KeyError:
			# Create a parser object with no name and the address of the parser
			parser_obj = Parser(None, parser_addr)
			self.parser_objs[parser_addr] = parser_obj

		parser_stack = self.peek_parserstack()
		parser_stack.push(parser_obj)
		if self.parser_decombinator:
			p_env = self.parser_decombinator.decompose_parser(parser_obj, self)
			self.set_parser_env(type(p_env).__name__ + " - " + str(p_env)) # TODO: pass this as data structure to frontend
		return parser_obj

	def return_from_perform_lowlevel_parse(self):
		parser_stack = self.peek_parserstack()
		parser_obj = parser_stack.pop()
		# debug print here

	def enter_h_arena_malloc_raw(self, alloc_size):
		parser_obj = self.peek_parser()
		parser_stack = self.peek_parserstack()
		# This is probably the slowest part of the code, or maybe the overhead adds up over many calls to h_arena_malloc_raw()
		if parser_obj is not None:
			# Caveat: parser_stack is assumed not to be None if we could get a parser_obj
			parser_obj.add_mem_use(parser_stack.parse_state, alloc_size)
		elif parser_stack is not None:
			#print("Allocation of " + str(alloc_size) + " bytes without a parser on the stack. (Happens before first call perform_lowlevel_parse to or after return from that call)")
			parser_stack.unclaimed_mem_use += alloc_size
		else:
			#print("Allocation of " + str(alloc_size) + " bytes without a parser stack. (This happens before and after parse)")
			self.unclaimed_mem_use += alloc_size

	def parse_virtual(self, parser_name):
		parser_obj = self.peek_parser()
		if parser_obj.name is None:
			parser_obj.name_parser(parser_name)
		#else:
			#print("Warning: parser already named! This is a bug. old name: %s, new name: %s" % (parser_obj.name, parser_name))

	#def return_from_parse_virtual(self, parse_state, result_valid):
	#	self.current_token_end = int(parse_state['input_stream']['index']) #TODO: maybe just have ParseVirtualRetBreakpoint extract the index and pass it into the function
	#	token_bounds = (self.current_token_start, self.current_token_end)

	def begin_new_token(self, index):
		token_bounds = (index, None)

		diff = (self.current_parse_depth+1) - len(self.starting_input_index)
		newindices = [None] * diff
		self.starting_input_index.extend(newindices)
		self.starting_input_index[self.current_parse_depth] = index
		#print(self.starting_input_index)

		#print("DEBUG: self.current_parse_depth:", self.current_parse_depth)
		#print("DEBUG: path after begin_new_token():", self.input_index_tree_path) # TODO: rename input_index_tree_path to next_node_path
		#print("DEBUG: tokens after begin_new_token():", self.input_stream_indices)
		#if len(self.input_stream_indices) <= self.current_parse_depth: # No tokens yet at this depth
		#	self.input_stream_indices.append([token_bounds]) # note: input_stream_indices is a list of lists, the "inner" list being bounds for tokens in a sequence
		#else: # There are already tokens at this depth (this happens when e.g. the preceding h_do_parse is the application of a h_sequence())
		#	self.input_stream_indices[self.current_parse_depth].append(token_bounds)

	# If a parser successfully returns a token, we take the list of token bounds at the current depth, check that the bounds corresponding to
	# the most recent parse result aren't filled in yet, and make note of state->input_stream.index
	def finish_token(self, index, hparseresult_addr):
		#print("DEBUG: finish_token()")
		#print("DEBUG: current parse depth:", self.current_parse_depth)
		#print("DEBUG: finished tokens:", self.input_stream_indices)
		#print("DEBUG: path:", self.input_index_tree_path)
		#print("DEBUG: path (truncated):", self.input_index_tree_path[0:self.current_parse_depth+1])
		#token = self.input_stream_indices[self.current_parse_depth][-1]

		token_addr = hparseresult_addr.cast(gdb.lookup_type("HParseResult").pointer())['ast']
		if token_addr:
			#print(HParseResult(hparseresult_addr).str_no_deref())
			#TODO: save parser directly, extracting the name at printing
			parsername = getattr(self.peek_parser(), 'name', '!!!UNNAMED!!!') # DEBUG
			self.input_token_map[int(token_addr)] = {'start': self.starting_input_index[self.current_parse_depth], 'end': index, 'hparseresult': hex(hparseresult_addr), 'parser': parsername}
			#print(self.input_token_map[int(token_addr)])
			#print(hex(token_addr))
		#print("DEBUG: current parse depth:", self.current_parse_depth)
		#print("DEBUG: finished tokens:", self.input_stream_indices)
		#print("DEBUG: path:", self.input_index_tree_path)
		#print([str(p) for p in self.peek_parserstack().p_stack])

	# Discard a single item in the "outer" list, which is a list of tokens at that depth in the call graph (one if the parser in the stack is not a sequence, multiple if it is)
	def discard_pending_tokens(self):
		#TODO: clean out unfinished tokens at the end
		#print("DEBUG: discard_pending_tokens()")
		#print("DEBUG: current parse depth:", self.current_parse_depth)
		#print("DEBUG: path:", self.input_index_tree_path)
		#print("DEBUG: tokens:", self.input_stream_indices)
		#tokens = self.input_stream_indices.pop()
		return

	def peek_parserstack(self):
		try:
			retval = self.parser_stacks[-1]
		except IndexError:
			retval = None
		return retval

	def peek_parser(self):
		try:
			retval = self.peek_parserstack().peek()
		except AttributeError:
			# print("Parser stack of stacks empty!")
			retval = None

		# retval will also be None when parser stack is empty (while parser stack of stacks isn't)
		return retval

	def parser_by_address(self, parser_addr):
		try:
			return self.parser_objs[int(parser_addr)]
		except KeyError:
			print("Parser with address " + str(hex(parser_addr)) + " not found!")

	def parsers_by_name(self, parser_name):
		results = [v for k,v in self.parser_objs.items() if v.name == parser_name]
		if len(results) > 0:
			return results
		else:
			return None

	def set_input_chunk(self, chunk):
		self.input_chunk = chunk

	def get_input_chunk(self):
		return self.input_chunk

	def set_parser_env(self, parser_env):
		self.current_parser_env = parser_env

	def get_parser_env(self):
		return self.current_parser_env

	def add_or_get_parser(self, parser_addr):
		try:
			parser_obj = self.parser_objs[int(parser_addr)]
		except KeyError:
			# Create a parser object with no name and the address of the parser
			parser_obj = Parser(None, int(parser_addr))
			self.parser_objs[int(parser_addr)] = parser_obj

		return parser_obj

	def get_parser_top_per_arena_mem(self):
		return sorted(self.parser_objs.values(), key=Parser.get_arenamax, reverse=True)[0]

	def get_parser_top_total_arena_mem(self):
		return sorted(self.parser_objs.values(), key=Parser.get_arenasum, reverse=True)[0]

	def get_avg_mem_use_per_arena(self):
		avg_mem_use = {}
		arena_counts = {}
		# Accumulate byte counts and counts of parsers using that arena
		for p in self.parser_objs.values():
			for arena,mem in p.bytes_used.items(): # TODO: state vs arena
				if arena in arena_counts.keys():
					arena_counts[arena] += 1
				else:
					arena_counts[arena] = 1

				if arena in avg_mem_use.keys():
					avg_mem_use[arena] += mem
				else:
					avg_mem_use[arena] = mem

		averages = {arena: mem/arena_counts[arena] for arena,mem in avg_mem_use.items()}

		return averages

	def dump_memory_stats(self):
		return [(p.name, hex(p.address), p.bytes_used) for p in self.parser_objs.values()]

	# TODO: better naming. this is for the AST view
	# This will only work as advertised if called from h_do_parse()
	def want_result_of(self, parser=None):
		# Check that:
		# - The relevant parser stack is active (h_packrat_parse hasn't been called again, or has returned since)
		# - A particular parser application on the stack has been selected
		# - The selected parser has just been popped off the stack:
		#	- the size of the stack is equal to the "stack index" (meaning it points to one element past the top of the stack, meaning the parser of interest has just been popped)
		#	- (Potential modification in the future: the selected application is on the top of the stack and the same parser is being applied)
		if self.ast_selected_stack and self.peek_parserstack() == self.ast_selected_stack and self.ast_stack_index and len(self.ast_selected_stack.p_stack) == self.ast_stack_index:
			return True
		else:
			return False

	def setup_ast_stack_index(self, stop_index):
		self.ast_selected_stack = self.peek_parserstack()
		# Indexing from the bottom of the stack
		if stop_index < 0:
			# Convert negative index to positive one
			self.ast_stack_index = abs(stop_index)-1
		# Indexing from the top of the stack
		elif stop_index > 0:
			# The "current" parser is not on the stack while we're in h_do_parse, as it gets pushed in perform_lowlevel_parse. Consider:
			# [0x5555555779f0] digit [current] <--- not on stack, for hammer-parse-step-to-parse-result, this is #0, or #-7

			#[0x55555557f1b0] (Unnamed sequence) <--- self.ast_selected_stack.p_stack[5], #1 or #-6
			#[0x55555557f200] xrefs
			#[0x5555555800f0] (Unnamed sequence)
			#[0x555555580140] xr_td
			#[0x555555585670] (Unnamed choice)
			#[0x55555559bb90] (Unnamed ignoreseq) <--- self.ast_selected_stack.p_stack[0], #6 or #-1

			self.ast_stack_index = len(self.ast_selected_stack.p_stack)-stop_index
		# 0 has been passed in: Select current parser
		else:
			self.ast_stack_index = len(self.ast_selected_stack.p_stack)

	def clear_ast_stack_index(self):
		self.ast_stack_counter = None
		self.ast_selected_stack = None

	# TODO: get_avg_mem_use_all_arenas, get_total_mem_use

	def print_input_chunk(self, token, token_list, rows, start, end):
		#print("print_input_chunk(): start: ", start, "end: ", end) # DEBUG
		w = end-start
		if w == 0:
			print("all tokens passed to display are zero-width")
			return
		tokenmap_val_list = [dict(token, addr=hex(key)) for key, token in self.input_token_map.items() if (token['start'] >= start and token['start'] <= end)]
		numrows = min(rows, len(tokenmap_val_list))
		h = numrows
		charbuf_dynamic = []
		current_row = 0
		info_strings = []
		printing_index = self.printing_index
		for index, token in enumerate(self.tokens_from_previous_chunk):
			tokenmap_entry = self.input_token_map[int(token.address)]
			token_length = tokenmap_entry['end'] - tokenmap_entry['start']
			start_col = max(tokenmap_entry['start'] - start, 0)
			end_col = min(tokenmap_entry['end'] - start, w)
			if tokenmap_entry['end'] < start: #TODO: better control flow here, deletion branch could be merged
				del self.tokens_from_previous_chunk[index]
				continue
			charbuf_dynamic.append([' '] * w)
			charbuf_dynamic[current_row][start_col:end_col] = ['X'] * min(token_length, end_col-start_col) # TODO: charbuf_dynamic[-1]
			hparseresult_addr = tokenmap_entry['hparseresult']
			current_row += 1

			try:
				hpr = HParseResult(int(hparseresult_addr, 16))
				if token.address != hpr.ast.address:
					raise RuntimeError("confused the info strings")
				info_strings.append(" ".join([hpr.str_no_deref(), str(tokenmap_entry)]))
			except Exception as err:
				print("Got exception when validating tokeninfo: ", err)
				info_strings.append(" ".join(["invalid", str(tokenmap_entry)]))

			if tokenmap_entry['end'] <= end:
				del self.tokens_from_previous_chunk[index] # TODO: Maybe it's a bad idea to modify a container you're iterating on


		while printing_index < len(token_list):
			token = token_list[printing_index]
			tokenmap_entry = self.input_token_map[int(token.address)]
			if tokenmap_entry['start'] > end:
				break
			token_length = tokenmap_entry['end'] - tokenmap_entry['start']
			start_col = max(tokenmap_entry['start'] - start, 0)
			end_col = min(tokenmap_entry['end'] - start, w)
			charbuf_dynamic.append([' '] * w)
			charbuf_dynamic[current_row][start_col:end_col] = ['X'] * min(token_length, end_col-start_col) # TODO: charbuf_dynamic[-1]
			hparseresult_addr = tokenmap_entry['hparseresult']
			current_row += 1

			try:
				hpr = HParseResult(int(hparseresult_addr, 16))
				info_strings.append(" ".join([hpr.str_no_deref(), str(tokenmap_entry)]))
				if token.address != hpr.ast.address:
					raise RuntimeError("confused the info strings")
			except Exception as a: #DEBUG
				print(a, "type:", type(a))
				print(traceback.format_exc())
				gdb.execute("list")
				gdb.execute("backtrace")
				info_strings.append(" ".join(["invalid", str(tokenmap_entry)]))

			if tokenmap_entry['end'] > end:
				self.tokens_from_previous_chunk.append(token)

			printing_index += 1

		self.printing_index = printing_index
		info_dict = dict(enumerate(info_strings))
		charbufrows_token_debug = [ "".join(row) + "\ntokeninfo: " + info_dict.get(index, "no token here") for index, row in enumerate(charbuf_dynamic)]
		#charbufrows = ["".join(row) for row in charbuf_dynamic]
		#charbuf_final = "\n".join(charbufrows)
		charbuf_final = "\n".join(charbufrows_token_debug)
		print(charbuf_final)

	def print_input_map(self, token, parent_bounds=None):
		w = gdb.parameter("width")
		rows = gdb.parameter("height")
		if token is None:
			print("AST pointer passed to print_input_map() is NULL")
			return
		#print(self.input_token_map)
		addr = token.address
		#addr = token
		try:
			start = self.input_token_map[int(addr)]['start']
		except KeyError as ke:
			print(self.input_token_map)
			print(ke)
		end = self.input_token_map[int(addr)]['end']
		i = 0
		printing_tokens = [token]
		self.printing_index = 0
		while i < len(printing_tokens):
			tk = printing_tokens[i]
			if tk.token_type == HParsedToken.TT_SEQUENCE:
				#printing_tokens.extend(tk.children[::-1])
				printing_tokens.extend(tk.children) # No longer using this as a stack, so no reversing is needed
			i += 1
		#debug_printing_token_list = [token.str_no_deref() for token in printing_tokens]
		#print(debug_printing_token_list)
		fragmentlen = end-start
		inputfragment = (self.input_ptr+start).string('UTF-8','replace',fragmentlen)
		#encoded_inputfragment = inputfragment.replace("\n", "<0A>").replace("\r", "<0D>").replace(" ", "<20>").replace("\t", "<09>")
		encoded_inputfragment = inputfragment.replace("\n", ".").replace("\r", ".").replace("\t", ".")
		screens = fragmentlen // w
		last_chunk_width = fragmentlen % w
		for chunk in range(0, screens):
			inputfragment = (self.input_ptr + (start + w * chunk) ).string('UTF-8', 'replace', w)
			encoded_inputfragment = inputfragment.replace("\n", ".").replace("\r", ".").replace("\t", ".")
			print(encoded_inputfragment)
			#print("chunk: ", chunk) # DEBUG
			self.print_input_chunk(token, printing_tokens, rows, start + w * chunk, (start + w * chunk) + min(w, end-start))
		#print("last chunk width:", last_chunk_width) # DEBUG
		inputfragment = (self.input_ptr + (start + w * screens)).string('UTF-8', 'replace', last_chunk_width)
		encoded_inputfragment = inputfragment.replace("\n", ".").replace("\r", ".").replace("\t", ".")
		print(encoded_inputfragment)
		self.print_input_chunk(token, printing_tokens, rows, start + w * screens, (start + w * screens) + min(w, end-start))
		self.tokens_from_previous_chunk = []
		#TODO: clean up self.starting_input_index when when h_parse returns

		#if rec_depth == 0:
		#	print(encoded_inputfragment)
		#	self.print_input_chunk(token, rows, start, start+min(w, end-start))
		#if token.children:
			#for child in token.children:
				#print("print_input_map recursion:", rec_depth+1)
			#	self.print_input_map(child,rec_depth+1,(start,end))

	# TODO: better name pending
	def get_current_stack_depth(self):
		return self.current_parse_depth # TODO: what's the difference between this and ast_stack_index?

top_level_parse = TopLevelParse()
